# frozen_string_literal: true
module Modelur
# Effect module provides a classes and static methods to use [algebraic effects](https://github.com/digital-fabric/affect).
  module Effect
    module_function

    # Implements an effects context. The effect context define [effect handlers](https://www.eff-lang.org/handlers-tutorial.pdf)
    # which perform the actual effects.
    class Context
      def initialize(**handlers, &block)
        @handlers = handlers || (block && { nil => block })
      end

      attr_reader :handlers

      # Add a new effect handler. If the handler for the effect already exist, it will be overridden.
      # @example
      #  context = Effect::Context.new({})
      #
      #  context.add_handler(effect: :hello) { puts "Hello"}
      #  context.add_handler(effect: :world, handler: ->{ puts " world" })
      #  context.add_handler(EffectHandler.new(:exclamation) {puts "!"})
      #  context.capture{
      #   Effect.perform(:hello)
      #   Effect.perform(:world)
      #   Effect.perform(:exclamation)
      #  }
      def add_handler(effect: nil, handler: nil, &block)
        effect ||= handler.respond_to?(:effect) ? handler.effect : nil
        @handlers[effect] = handler || block
      end

      def handler_proc
        proc { |effect, *args| handle(effect, *args) }
      end

      # Perform the given effect. This method is called by the static method {Effect#perform} and should not be called
      # from context objects.
      # @param effect [Object] the effect to perform. It is usually a {Symbol}, but can be any object
      def perform(effect, *args, &block)
        handler = find_handler(effect)
        if handler
          call_handler(handler, effect, *args, &block)
        elsif @parent
          @parent.perform(effect, *args, &block)
        else
          raise "No handler found for #{effect.inspect}"
        end
      end

      def find_handler(effect)
        @handlers[effect] || @handlers[effect.class] || @handlers[nil]
      end

      def call_handler(handler, effect, *args, &block)
        if handler.arity == 0
          handler.call(&block)
        elsif args.empty?
          handler.call(effect, &block)
        else
          handler.call(*args, &block)
        end
      end

      @@current = nil
      def self.current
        @@current
      end

      def capture
        @parent, @@current = @@current, self
        catch(:escape) { yield }
      ensure
        @@current = @parent
      end

      # Perform some actions at the end of the capture block
      # This method is empty and can be overloaded by classes that inherit from the {Context}
      # to perform some additional work at the end of the capture block
      def finalize; end

      def escape(value = nil)
        throw :escape, (block_given? ? yield : value)
      end
    end

    def capture(*args, &block)
      block, handlers = block_and_handlers_from_args(*args, &block)
      handlers = { nil => handlers } if handlers.is_a?(Proc)
      Context.new(handlers).capture(&block)
    end

    def block_and_handlers_from_args(*args, &block)
      case args.size
      when 1 then block ? [block, args.first] : [args.first, nil]
      when 2 then args
      else [block, nil]
      end
    end

    # Perform the effect. This method sends the effect to the first {Context} object that can handle the effect. The
    # {Context} object than invoke the {Context#perform} method that invokes the right handler or sends it up to other
    # {Context} object.
    # @param effect [Object] the effect to perform. It is usually a {Symbol}, but can be any object
    # @example
    #  def mul(x, y)
    #    Effect.perform(:log, "Multiply #{x} and #{y}")
    #  end
    #  Effect.capture(log: { |message| puts message} ) do
    #     mul(2, 3)
    #  end
    #  # prints "Multiply 2 and 3"
    #  logger = []
    #  Effect.capture(log: {|message| logger << message}) do
    #    mul(2, 3)
    #  end
    #  # logger = ["Multiply 2 and 3"]
    def perform(effect, *args, &block)
      unless (ctx = Context.current)
        raise "perform called outside capture block with #{effect}"
      end

      ctx.perform(effect, *args, &block)
    end

    def escape(value = nil, &block)
      unless (ctx = Context.current)
        raise 'escape called outside capture block'
      end

      ctx.escape(value, &block)
    end
  end
end
